﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

// type definitions for ease of use

namespace HeuristicLab.EvolutionaryTracking {
  /// <summary>
  /// Population diversity analyzer
  /// </summary>
  [Item("SymbolicExpressionTreeEvolvabilityAnalyzer", "An operator that measures relative improvements obtained by genetic operators.")]
  [StorableClass]
  public sealed class SymbolicExpressionTreeEvolvabilityAnalyzer : SingleSuccessorOperator, IAnalyzer {
    #region Parameter names
    private const string SymbolicExpressionTreeParameterName = "SymbolicExpressionTree";
    private const string UpdateIntervalParameterName = "UpdateInterval";
    private const string UpdateCounterParameterName = "UpdateCounter";
    private const string ResultsParameterName = "Results";
    private const string PopulationGraphParameterName = "PopulationGraph";
    private const string PopulationSizeParameterName = "PopulationSize";
    private const string ElitesParameterName = "Elites";
    #endregion

    #region Parameter properties
    public ValueParameter<IntValue> UpdateIntervalParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateIntervalParameterName]; }
    }
    public ValueParameter<IntValue> UpdateCounterParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateCounterParameterName]; }
    }
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public IScopeTreeLookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (IScopeTreeLookupParameter<ISymbolicExpressionTree>)Parameters[SymbolicExpressionTreeParameterName]; }
    }
    public LookupParameter<IntValue> ElitesParameter {
      get { return (LookupParameter<IntValue>)Parameters[ElitesParameterName]; }
    }
    public LookupParameter<IntValue> PopulationSizeParameter {
      get { return (LookupParameter<IntValue>)Parameters[PopulationSizeParameterName]; }
    }
    #endregion

    #region Parameters
    public IntValue UpdateInterval { get { return UpdateIntervalParameter.Value; } }
    public IntValue UpdateCounter { get { return UpdateCounterParameter.Value; } }
    public ResultCollection Results { get { return ResultsParameter.ActualValue; } }
    public IntValue PopulationSize { get { return PopulationSizeParameter.ActualValue; } }
    public IntValue Elites { get { return ElitesParameter.ActualValue; } }
    #endregion

    [StorableConstructor]
    private SymbolicExpressionTreeEvolvabilityAnalyzer(bool deserializing) : base(deserializing) { }
    private SymbolicExpressionTreeEvolvabilityAnalyzer(SymbolicExpressionTreeEvolvabilityAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicExpressionTreeEvolvabilityAnalyzer(this, cloner);
    }
    public SymbolicExpressionTreeEvolvabilityAnalyzer() {
      // add parameters
      Parameters.Add(new ScopeTreeLookupParameter<ISymbolicExpressionTree>(SymbolicExpressionTreeParameterName, "The symbolic expression trees to analyze."));
      Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
      Parameters.Add(new ValueLookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<IntValue>(PopulationSizeParameterName, "The population size."));
      Parameters.Add(new LookupParameter<IntValue>(ElitesParameterName, "Number of elites in the population."));

      UpdateCounterParameter.Hidden = true;
      UpdateIntervalParameter.Hidden = true;
    }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() { }

    #region IStatefulItem members
    public override void InitializeState() {
      base.InitializeState();
      UpdateCounter.Value = 0;
    }

    public override void ClearState() {
      base.ClearState();
      UpdateCounter.Value = 0;
    }
    #endregion

    public override IOperation Apply() {
      UpdateCounter.Value++;

      if (UpdateCounter.Value == UpdateInterval.Value) {
        UpdateCounter.Value = 0;

        if (!Results.ContainsKey(PopulationGraphParameterName)) return base.Apply();
        var graph = Results[PopulationGraphParameterName].Value as SymbolicExpressionTreeGenealogyGraph;
        if (graph == null) return base.Apply();

        if (graph.Nodes.Count < PopulationSize.Value) return base.Apply();

        var trees = SymbolicExpressionTreeParameter.ActualValue.ToList();

        var graphNodes = trees.Select(x => graph.GetGraphNodes(x).Last());

        int count = 0;

        foreach (var node in graphNodes) {
          double parentQuality = 0.0;
          if (node.InEdges == null) continue;
          switch (node.InEdges.Count) {
            case 2: // crossover
              parentQuality = node.InEdges.Max(x => (x.Source as SymbolicExpressionGenealogyGraphNode).Quality);
              break;
            case 1: // mutation
              parentQuality = (node.InEdges[0].Source as SymbolicExpressionGenealogyGraphNode).Quality;
              break;
          }

          if (node.Quality > parentQuality) ++count;
        }
        double reproductiveSuccess = (double)count / (PopulationSize.Value - Elites.Value);

        if (!Results.ContainsKey("RelativeReproductiveSuccess"))
          Results.Add(new Result("RelativeReproductiveSuccess", new DataTable("Reproductive success")));

        var relativeReproductiveSuccessTable = Results["RelativeReproductiveSuccess"].Value as DataTable;
        if (relativeReproductiveSuccessTable == null) return base.Apply();
        if (!relativeReproductiveSuccessTable.Rows.ContainsKey("Reproductive Success"))
          relativeReproductiveSuccessTable.Rows.Add(new DataRow("Reproductive Success") { VisualProperties = { StartIndexZero = true } });
        relativeReproductiveSuccessTable.Rows["Reproductive Success"].Values.Add(reproductiveSuccess);

        if (!Results.ContainsKey("RelativeProportionOfLeafNodes"))
          Results.Add(new Result("RelativeProportionOfLeafNodes", new DataTable("Proportion of leaf nodes")));
        var relativeProportionOfLeafNodesTable = Results["RelativeProportionOfLeafNodes"].Value as DataTable;
        if (!relativeProportionOfLeafNodesTable.Rows.ContainsKey("Percent of leaf nodes"))
          relativeProportionOfLeafNodesTable.Rows.Add(new DataRow("Percent of leaf nodes") { VisualProperties = { StartIndexZero = true } });
        if (!relativeProportionOfLeafNodesTable.Rows.ContainsKey("Average number of leafs"))
          relativeProportionOfLeafNodesTable.Rows.Add(new DataRow("Average number of leafs") { VisualProperties = { StartIndexZero = true } });

        double totalNodes = trees.Sum(x => x.Length);
        double totalLeafs = trees.Sum(t => t.IterateNodesPostfix().Count(n => n.SubtreeCount == 0));

        relativeProportionOfLeafNodesTable.Rows["Percent of leaf nodes"].Values.Add(totalLeafs / totalNodes);
        relativeProportionOfLeafNodesTable.Rows["Average number of leafs"].Values.Add(totalLeafs / trees.Count);
      }
      return base.Apply();
    }

    public bool EnabledByDefault { get { return true; } }
  }
}
