﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionaryTracking {
  [StorableClass]
  public class Vertex : IVertex {
    public string Id { get; private set; }
    public string Label { get; set; }
    public double Weight { get; set; }
    public List<Arc> InEdges { get; private set; }
    public List<Arc> OutEdges { get; private set; }
    private object content;
    public object Content {
      get { return content; }
      set {
        if (value == null)
          throw new ArgumentException("Node content cannot be null.");
        content = value;
      }
    }

    public Vertex() {
      Id = Guid.NewGuid().ToString();
    }

    public Vertex(Vertex node) {
      Id = Guid.NewGuid().ToString();
      Label = node.Label;
      Content = node.Content;
      InEdges = new List<Arc>(node.InEdges);
      OutEdges = new List<Arc>(node.OutEdges);
    }

    public int InDegree { get { return InEdges == null ? 0 : InEdges.Count; } }
    public int OutDegree { get { return OutEdges == null ? 0 : OutEdges.Count; } }
    public int Degree { get { return InDegree + OutDegree; } }

    /// <summary>
    /// Performs an upward-breath-traversal of the graph using the nodes InEdges
    /// </summary>
    /// <returns>All the ancestors of the current node</returns>
    public IEnumerable<IVertex> Ancestors() {
      // for performance, we use a hashset for lookup and a list for iteration
      var nodes = new HashSet<IVertex> { this };
      var list = new List<IVertex> { this };
      int i = 0;
      while (i != list.Count) {
        if (list[i].InEdges != null) {
          foreach (var e in list[i].InEdges) {
            if (nodes.Contains(e.Source)) continue;
            nodes.Add(e.Source);
            list.Add(e.Source);
          }
        }
        ++i;
      }
      return list;
    }

    /// <summary>
    /// Performs a downward-breath-traversal of the genealogy graph using the nodes OutEdges
    /// </summary>
    /// <returns>All the descendants of the current node</returns>
    public IEnumerable<IVertex> Descendants() {
      var nodes = new HashSet<IVertex> { this };
      var list = new List<IVertex> { this };
      int i = 0;
      while (i != list.Count) {
        if (list[i].OutEdges != null) {
          foreach (var e in list[i].OutEdges) {
            if (nodes.Contains(e.Target)) continue;
            nodes.Add(e.Target);
            list.Add(e.Target);
          }
        }
        ++i;
      }
      return list;
    }

    // this method can be used to add arcs towards targets that are not visible in the graph 
    // (they do not appear in the nodes Dictionary). It is the programmers responsibility to 
    // enforce the correct and desired behavior.
    public void AddForwardArc(IVertex target, double weight = 0.0, object data = null) {
      var e = new Arc { Source = this, Target = target, Data = data, Weight = weight };
      if (OutEdges == null) OutEdges = new List<Arc> { e };
      else OutEdges.Add(e);
    }
    public void AddForwardArc(Arc arc) {
      if (arc.Source == null) { arc.Source = this; }
      if (arc.Source != this) { throw new Exception("AddForwardArc: Source should be equal to this."); }
      if (OutEdges == null) { OutEdges = new List<Arc> { arc }; } else { OutEdges.Add(arc); }
    }

    public void AddReverseArc(Arc arc) {
      if (arc.Target == null) { arc.Target = this; }
      if (arc.Target != this) { throw new Exception("AddReverseArc: Target should be equal to this."); };
      if (InEdges == null) { InEdges = new List<Arc> { arc }; } else { InEdges.Add(arc); }
    }


    public void AddReverseArc(IVertex source, double weight = 0.0, object data = null) {
      var e = new Arc { Source = source, Target = this, Data = data, Weight = weight };
      if (InEdges == null) InEdges = new List<Arc> { e };
      else InEdges.Add(e);
    }
  }
}
